"Python3 document categorization library, syllexis method."

##  SPANSETS : :
##
##  Python3 data & routines to supply to toccata test harness s/w :
##  Copyright 2017, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 27/01/2017
##  last revision : 23/02/2017


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys
##import  random


if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
    

from globdefs import *
from glib import  add2dic, numerize
from tocclib import  bastats
from tocclib import  corsums, corrval
from tocclib import  midvalue, ranx
from tocclib import  fourfold, tetracor
from tocclib import  slugger, slugging
from tocclib import  g2, hedgesg


##  global vars :
Showoff = 0  ##  >0 for debugging o/p
Gold = 5.0**0.5*0.5+0.5  ##  good approximation to golden number
Nulltok = ".d0t."
Maxtoks = 48 * 1024  ##  slow with large files, so truncate (?)
Iced = type(frozenset([1]))
##  print(Iced)


##  functions first, then class(es) :

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  TOPTOKS :
def toptoks (freqtab,h=144):
    "returns set of high-frequency tokens from frequency dict."

    v = len(freqtab)  ##  vocsize
    sortlist = []
    for k in freqtab:
        f = freqtab[k]
        if f > 1:  ## always skip hapaxes
            sortlist.append((f,k))
    sortlist.sort(reverse=True)

    return  sortlist[:h]


##  ADDSNIPS :
def addsnips (d,snipsize):
    "adds snippet data to text d if not already present."

    d.snippets = 0
    d.snipsets = []  ##  list of sets
    minsnip = int(round(snipsize / Gold))
    for p in range(0,d.size,snipsize):
        slab = d.toklist[p:p+snipsize]
        if len(slab) >= minsnip or p==0:
            d.snipsets.append(set(slab))
            d.snippets += 1

    return  ##  side-effect on d.


##  G2SCORE :
def g2score (n,frex,nulrates):
    "quick & dirty g-squared score from frex & base rates."

    ##  n should be sum(frex).
    assert len(frex) == len(nulrates)
    g2 = 0.0
    for j,f in enumerate(frex):
        if frex[j] > 0:
            e = n * nulrates[j]
            gx = frex[j] * math.log(frex[j]/e)
            ##  e must never be zero !
            g2 += gx

    return  g2


##  FREQLEN :
def freqlen (gtab):
    "orders gtab by frequency then length."

    sortlist = []
    for g in gtab:
        f = gtab[g]
        nc = len(' '.join(g))
        if f > 2:  ##  hapaxes & dislegomena dropped
            sortlist.append((f,len(g),nc,g))

    sortlist.sort(reverse=True)
    return  sortlist
    ##  format : (freq,toksize,charsize,gram).


##  FINDLOCS :
def findlocs (toklist,item):
    "returns locations of item in token list."

    locs = []
    p = 0
    nt = len(toklist)
    while p < nt:
        try:
            p = toklist.index(item,p)
            locs.append(p)
            p += 1
        except ValueError as xx:
            break

    return  locs


##  TOKSLABS :
def tokslabs (toklist,spansize,stopmode,spanmode,topstops):
    "collects slabsets from a token list."

    if spanmode > 0:  ##  tuples wanted
        return  toktups(toklist,spansize,stopmode,topstops)
    ##  iced-sets wanted; compute stepsize :
    nt = len(toklist)
    stepsize = int(math.log10(nt))
    if stepsize < 2:
        stepsize = 1  ##  avoid zero !
    elastab = {}
    for p in range(0,nt-spansize,stepsize):
        slab = toklist[p:p+spansize]
        if stopmode == 0:
            slabset = set(slab) & topstops
        elif stopmode == 1:
            slabset = set(slab) - topstops
        else:  ##  free choice !
            slabset = set(slab)
        fs = frozenset(slabset)
        if fs in elastab:
            elastab[fs] += 1
        else:
            elastab[fs] = 1

    return  elastab


##  TOKTUPS :
def toktups (toklist,spansize,stopmode,topstops):
    "collects slab-tuples from a token list (spanmode > 0)."

    ##  compute stepsize :
    nt = len(toklist)
    stepsize = int(math.log10(nt))
    if stepsize < 2:
        stepsize = 1  ##  avoid zero !
    elastab = {}
    for p in range(0,nt-spansize,stepsize):
        slab = toklist[p:p+spansize]
        if stopmode == 0:
            stup = [t for t in slab if t in topstops]
        elif stopmode == 1:
            stup = [t for t in slab if not (t in topstops)]
        else:  ##  free choice !
            stup = slab
        fs = tuple(stup)
        if fs in elastab:
            elastab[fs] += 1
        else:
            elastab[fs] = 1

    return  elastab


##  BINORMS :
def binorms (f1,n1,freq,tote):
    "Forman's (2003) binormal separation from local freq & global."
    ##  f1,n1 are part of freq,tote; frex of subsample.

    if freq < 1 or tote < 1 or n1 < 1 or n1 >= tote:
        return 0
    baserate = freq / tote
    x1 = math.sqrt(n1)
    p1 = (f1+baserate*x1) / (n1+x1)
    ##  compute frex in unselected class :
    f2 = freq - f1
    n2 = tote - n1
    x2 = math.sqrt(n2)
    p2 = (f2+baserate*x2) / (n2+x2)
    ##  print(p1,p2)
    ##  print(p1*(1-p1),p2*(1-p2))
    ##  rates have been attenuated; next compare z-scores :
    diff = p1 / math.sqrt(p1*(1-p1)) - p2 / math.sqrt(p2*(1-p2))

    return  diff

##print(binormsep(10,2000,20,12000))


####  DOCADAT :
class Docadat (object):
    ##  collection of data+methods to do document classification :
    ##  assumes access to 2 Sack() objects as args, doclist & paradat.
    ##  doclist attributes expected include:
    ##    classnum, freqtab, outcome, size, text, textlen, toklist,
    ##  paradat attributes expected include:
    ##    catlist, cats, paraline, testmeta, trainmet,

    def __init__ (self,paradat,doclist):
        "initialization method for Docadat objects."

        self.classes = paradat.cats
        self.catlist = paradat.catlist
        self.id = paradat.trainmet
        if hasattr(paradat,"testmeta"):
            self.id = self.id + '_' + paradat.testmeta
        self.alldocs = len(doclist)  ##  number of training-corpus documents

        return

    def loadpars (self,paradat,sep1=' ',sep2='='):
        "loads extra docalib parameters if any, provided in paraline."

        parline = paradat.paraline.lower().strip()
        self.pars = Sack(parline)
        parbits = parline.split(sep1)  ##  item separator defaults to space
        for pb in parbits:
            bits = pb.split(sep2)  ##  internal separator (e.g. "gramsize=4")
            if len(bits) == 2:
                v = numerize(bits[1])
                if type(v) == type(9) or type(v) == type(0.5):  ##  numeric
                    setattr(self.pars,bits[0],v)
                else:  ##  nonnumeric
                    setattr(self.pars,bits[0],bits[1])
                ##  no limits imposed !
        ##  check for suggested tops2use :
        tv = 0
        if hasattr(self.pars,"topvocs") and isinstance(self.pars.topvocs,int):
            self.topvocs = getattr(self.pars,"topvocs")
        else:
            self.topvocs = tv
        ##  tops2use won't use this if < 2 or > 256.
        tm = 0
        if hasattr(self.pars,"stopmode") and isinstance(self.pars.stopmode,int):
            self.stopmode = getattr(self.pars,"stopmode")
        else:
            self.stopmode = tm  ##  default
        self.stopmode = min(2,max(0,self.stopmode))  ##  enforce limits
        ##  0 omit stopwords; 1 omit all except stopwords;
        ##  2 keep all words. (!?)
        ss = 144  ##  default snipsize
        if hasattr(self.pars,"snipsize") and isinstance(self.pars.snipsize,int):
            self.pars.snipsize = numerize(getattr(self.pars,"snipsize",ss))
        else:
            self.pars.snipsize = ss  ##  may still be altered if unrealistic
        self.spansize = 3  ##  seems reasonable default
        if hasattr(self.pars,"spansize") and isinstance(self.pars.spansize,int):
            self.spansize = self.pars.spansize
        self.spansize = min(10,max(1,self.spansize))
        ##  1 is possible, though not very interesting.
        self.keymode = 5  ##  default is scaled ratediff
        if hasattr(self.pars,"keymode") and isinstance(self.pars.keymode,int):
            self.keymode = self.pars.keymode
        self.keymode = min(5,max(0,self.keymode))
        ##  0 means pick vars just on frequency in class.
        self.spanmode = 1  ##  default is tuples
        if hasattr(self.pars,"spanmode") and isinstance(self.pars.spanmode,int):
            self.spanmode = self.pars.spanmode
        self.spanmode = min(1,max(0,self.spanmode))
        ##  0 means sets; 1 means tuples.
        
        return  ##  doesn't crash if paradat.paraline empty

    def modprep (self,paradat):
        "does once-only overall tasks."

        ##  check that snipsize is reasonable :
        ns = paradat.tottoks / self.pars.snipsize
        if ns < 64 or ns > 8192:
            self.pars.snipsize = int(paradat.tottoks ** 0.5)
            print("snipsize altered to",self.pars.snipsize)
        else:
            print("snipsize =",self.pars.snipsize)
        ##  attach snippets to docs :
        totsnips = 0
        minsnip = int(round(self.pars.snipsize / Gold))
        for d in paradat.doclist:  ##  deals with entire sample
            d.toklist = d.toklist[:Maxtoks]  ##  enforce size limit
            ##  d.freqtab already computed; handy for vocsizes
            d.snippets = 0
            d.snipsets = []  ##  list of sets
            for p in range(0,d.size,self.pars.snipsize):
                slab = d.toklist[p:p+self.pars.snipsize]
                if len(slab) >= minsnip or p==0:
                    d.snipsets.append(set(slab))
                    d.snippets += 1
                    totsnips += 1
        print("Number of snippets =",totsnips)
        paradat.totsnips = totsnips  ##  does belong to whole corpus
        ##  now find high-freq tokens :
        tokfrex = {}
        for d in paradat.doclist:
            for snip in d.snipsets:
                for t in snip:
                    if t in tokfrex:
                        tokfrex[t] += 1
                    else:
                        tokfrex[t] = 1
            d.snippets = len(d.snipsets)  ##  record number of snippets in d
        paradat.highfrex = toptoks(tokfrex,1024)
        ##  compile overall tables once only, at start :
        self.fullfrex = {}  ##  overall training-corpus frequency table
        ##  compile overall freq table :
        for d in paradat.doclist:
            add2dic(self.fullfrex,d.freqtab)
        self.docs = len(paradat.doclist)
        self.vocsize = len(self.fullfrex)  ##  vocabulary size
        ##  used to decide number of common terms to keep.
        if self.topvocs >= 2 and self.topvocs < 256:  ##  sensible from user
            self.tops2use = self.topvocs
        else:  ##  topvocs not given or looks wild
            hc = math.log1p(self.vocsize) / math.log1p(paradat.tottoks)
            ##  Herdan's C.
            if hc < 0.4444:  ##  probably tags rather than words
                self.tops2use = self.vocsize - 1  ##  almost all
                print("hc",hc)
                self.tagmode = 1  ##  save for later
            else:  ##  assume word tokens
                self.tops2use = round(self.vocsize**0.5)
                self.tagmode = 0
        if self.tops2use > 256:  ##  still don't go unreasonably large
            self.tops2use = 256
        ##  otherwise stick with heuristic/default.
        self.toplist = []
        for hf in paradat.highfrex:
            if len(hf[1]) < 12:  ##  skip very long words
                self.toplist.append(hf)
            if len(self.toplist) >= self.tops2use:
                break
        ##  print('#',len(self.toplist))
        self.toptoks = [x[1] for x in self.toplist]  ##  without frex
        self.tops = set(self.toptoks)
        ##  high-frequency vocs, in effect 'stop-words'.
        ##  next collect slabsets :
        for d in paradat.doclist:
            d.elastab = tokslabs(d.toklist,self.spansize,self.stopmode,
                                 self.spanmode,self.tops)
            d.elastot = sum(d.elastab.values())
            if d.dnum < 4:
                print(d.dnum,file=paradat.fo)
                print(d.toklist[:99],file=paradat.fo)
                print(d.elastab,file=paradat.fo)
        
        return
    
    def makemods (self,paradat,doclist):
        "builds classification models for each category."
        ##  doclist will be training data.

        self.modlist = []  ##  list of category models
        self.modeltab = {}  ##  dictionary of models, if lookup by name needed
        self.name2num = {}
        self.fulltab = {}
        self.fulltote = 0
        ##  make models :
        cn = 0
        for c in self.catlist:  ##  relies on consistent order of catlist
            m = self.makemod(c,cn,paradat,doclist)
            self.modlist.append(m)
            self.modeltab[c] = m  ##  also point from dict keyed by catname
            self.name2num[c] = cn
            cn += 1  ##  keep track of position in model list        
        self.mods = len(self.modlist)
        assert  self.mods == self.classes

        templist = [(self.fulltab[x],x) for x in self.fulltab if self.fulltab[x] > 2]
        self.toplist = sorted(templist,reverse=True)[:self.tops2use]
        print(self.toplist,file=paradat.fo)
        self.topset = set([item[1] for item in self.toplist])
        self.allkeys = self.pickkeys(paradat,doclist)
        
        return

    def makemod (self,c,cn,paradat,doclist):
        "builds model info for a single category."

        m = Sack(c)  ##  gathered into a Sack() object
        m.docs = 0
        m.tottoks = 0
        m.catname = c
        m.classnum = cn
        m.spansets = 0
        m.toplist = []
        m.elastabs = {}
        m.elastots = 0
        m.vocfrex = {}  ##  only used to compute vocsize (?)
        for d in doclist:
            if d.outcome == c:
                m.docs += 1
                ##  safer for this library not2 cause side-effects in caller !
                m.tottoks += d.size
                m.spansets += len(d.elastab)
                add2dic(m.elastabs,d.elastab)
                m.elastots += d.elastot
                add2dic(self.fulltab,d.elastab)
                self.fulltote += d.elastot
                add2dic(m.vocfrex,d.freqtab)
        m.vocsize = len(m.vocfrex)
        ##  count singletons :
        haps = 0
        for e in m.elastabs:
            if m.elastabs[e] == 1:
                haps += 1
        m.haps = haps
        m.haprate = max(0.5,m.haps) / max(1,len(m.elastabs))
        print(c,self.fulltote,len(self.fulltab),file=paradat.fo)
        print(m.elastots,m.elastots/(max(1,len(m.elastabs))),file=paradat.fo)
        print(m.haps,m.haprate,file=paradat.fo)
        
        return  m

    def pickkeys (self,paradat,doclist):
        "selects useful subset of variables."

        invgold = 1.0 / Gold
        ##  collect keys from each class :
        for cn,catname in enumerate(self.catlist):
            kset = set()
            keylist = []
            m = self.modlist[cn]
            if m.docs <= 2:
                dmin = 1
            else:
                dmin = 2  ##  docfreq cutoff
            ##  decide number of keys to keep :
            ##  keep = round(math.sqrt(m.elastots-m.haps))
            evox = len(m.elastabs)
            k1 = round(m.vocsize**0.5)
            k0 = round(evox**invgold)
            ##  k0 = round(evox**0.5 * Gold)  ##  need rather more
            ##  if evox*Gold > m.vocsize:  ##  probably tags rather than words
            if self.tagmode > 0:  ##  assume tags, not word tokens
                keep = k0
                if self.spansize < 2:  ##  single tags
                    keep = int(m.vocsize / self.classes)
                else:
                    k0 = evox ** 0.5
                    mvox = min(144,m.vocsize**0.5)
                    keep = round(k0 + mvox + m.docs*Gold)
            else:  ##  probably word tokens
                keep = k1
            ##  print(evox,m.vocsize,keep,k0,k1,self.tops2use,m.elastots,m.haps)
            ##  if keep < self.tops2use:
                ##  keep = self.tops2use
            ##  keep = int(len(m.elastabs)**0.5)
            ##  keep = int(m.haps**0.5)
            ##  keep = int(m.elastots**0.5)
            ##  print(len(m.elastabs),m.elastots,m.haps,keep,m.vocsize,
            ##        int(m.vocsize**0.5))
            ##  keep = self.tops2use
            print(cn,catname,"keeping",keep,file=paradat.fo)
            rootdiv = m.docs ** 0.5 / 100.0
            for fs in m.elastabs:
                freq = m.elastabs[fs]
                if freq <= 2:
                    continue  ##  too rare, omit
                ##  compute document frequency :
                docfreq = 0
                ##  pvec = [] ; nvec = []
                for d in doclist:
                    ##  xf = d.elastab.get(fs,0)
                    ##  spanrate = xf / max(1,d.elastot)
                    if d.outcome != catname:
                        ##  nvec.append(spanrate)
                        continue
                    ##  pvec.append(spanrate)
                    if fs in d.elastab:
                        docfreq += 1
                if docfreq <= dmin:
                    continue  ##  too rare to keep
                ##  spanrate for hedgesg not worth the time.
                ##  compute keyscore of fs :
                f1 = m.elastabs[fs] ; n1 = m.elastots
                fx = self.fulltab[fs] - f1
                nx = self.fulltote - n1
                ##slugrate = slugger([f1/n1,fx/nx])  ##no diff! (2017/2/15)
                slugrate = slugging(f1,n1,fx,nx)
                if slugrate[0] < slugrate[1]:
                    continue  ##  don't save negative keys
                if self.keymode == 0:
                    keyscore = f1  ##  just pick on frequency !
                elif self.keymode in [1,4]:
                    keyscore = slugrate[0] - slugrate[1]
                    ##  simple rate diff.
                    if self.keymode == 4:
                        keyscore *= math.sqrt(f1)  ##  scaled rate-diff
                elif self.keymode == 2:
                    keyscore = binorms(f1,n1,self.fulltab[fs],self.fulltote)
                    ##  bi-normal separation, Forman (2003).
                elif self.keymode == 3:  ##  roar is mode 3
                    keyscore = (1-slugrate[1]) / (1-slugrate[0])  ##  roar
                else:  ##  5+ is ratediff by root docfreq, scaled
                    keyscore = (slugrate[0]-slugrate[1])*(docfreq**0.5/rootdiv)
                ##  roar seems okay; g2 & hedgesg probably not worth it.
                if self.spanmode == 0:
                    keylist.append((keyscore,docfreq,f1,n1,sorted(fs)))
                else:
                    keylist.append((keyscore,docfreq,f1,n1,fs))
                ##  last item 2b shown as list, if icedset.
            keylist.sort(reverse=True)
            m.keylist = keylist[:keep]  ##  :tops2use ?                

        ## aggregate :
        allkeys = set()
        for cn,catname in enumerate(self.catlist):
            m = self.modlist[cn]
            for item in m.keylist:
                if self.spanmode == 0:
                    allkeys.add(frozenset(item[-1]))  ## last item as icedset
                else:  ##  last item as tuple
                    allkeys.add(item[-1])

        return  allkeys  ##  set of icedsets or tuples

    def showmod (self,cn,fo=sys.stdout):
        "shows info related to model of category number cn."

        m = self.modlist[cn]
        assert cn == m.classnum
        print("\ncategory:",m.catname,'(',cn,')',file=fo)
        print("tokens =",m.tottoks,file=fo)
        print("spansets =",len(m.elastabs),file=fo)
        print("vocsize =",m.vocsize,file=fo)
        mf = midvalue(m.elastabs.values())
        frex = sorted(set(m.elastabs.values()),reverse=True)
        templist = [(m.elastabs[x],x) for x in m.elastabs if m.elastabs[x] > 2]
        
        print(mf,m.elastots,file=fo)
        print(frex,file=fo)
        tt = sorted(templist,reverse=True)
        ##  for j,item in enumerate(tt):
            ##  print(j,item,round(100.0*item[0]/m.elastots,4),file=fo)
        for i,item in enumerate(m.keylist):
            print(i,item,file=fo)
        
        return

    def showmods (self,paradat,fo=sys.stdout):
        "shows models of all categories."

        print(paradat.dateline,file=fo)
        print("classes =",self.classes,file=fo)
        print(paradat.catlist,file=fo)
        print("docs =",paradat.docs,file=fo)
        print("tokens =",paradat.tottoks,file=fo)
        print("vocsize =",self.vocsize,file=fo)
        print("topterms =",self.tops2use,file=fo)
        print("keymode =",self.keymode,file=fo)
        print("stopmode =",self.stopmode,file=fo)
        print("spanmode =",self.spanmode,file=fo)
        print("spansize =",self.spansize,file=fo)
        print("toplist :",self.toplist,file=fo)
        print("\ntoptoks :",self.toptoks,file=fo)
        print(self.tops,NL,file=fo)
        print(sorted(self.topset),NL,file=fo)
        print('',file=fo)
        print(len(self.allkeys),NL,self.allkeys,file=fo)
        print(NL,file=fo)

        for cn in range(self.classes):
            self.showmod(cn,fo)
            
        return

    def logbayes (self,cn,thisdoc,paradat):
        "naive bayes similarity of category number cn to given document."

        if not hasattr(thisdoc,"elastab"):
            thisdoc.elastab = tokslabs(thisdoc.toklist,self.spansize,
                                 self.stopmode,self.spanmode,self.tops)
            thisdoc.elastot = sum(thisdoc.elastab.values())
        m = self.modlist[cn]
        kset = set(thisdoc.elastab.keys())
        kset = kset | self.topset  ##  ensure globally frequent spansets
        ##  kset = self.allkeys
        divi = m.elastots + Gold  ##  overestimate
        logscore = 0
        for k in kset:
            x = thisdoc.elastab.get(k,1/Gold)
            y = m.elastabs.get(k,0)
            ##  if y < 1:
                ##  phat = m.haprate / (divi+m.haps)
            ##  else:
                ##  phat = y / divi
            phat = (y+0.125) / (divi)
            e = math.log(phat)
            logscore += e * x

        return  thisdoc.elastot / max(1,abs(logscore))
        ##  return  100/abs(logscore)

    def corrsim (self,cn,thisdoc,paradat):
        "correlational similarity of category number cn to given document."

        if not hasattr(thisdoc,"elastab"):
            thisdoc.elastab = tokslabs(thisdoc.toklist,self.spansize,
                                 self.stopmode,self.spanmode,self.tops)
            thisdoc.elastot = sum(thisdoc.elastab.values())     
        ##  kset = set(thisdoc.elastab.keys())
        ##  kset = kset | self.topset  ##  ensure globally frequent spansets
        kset = self.allkeys
        xvec = [] ; yvec = []
        for k in kset:
            x = thisdoc.elastab.get(k,0)
            y = self.modlist[cn].elastabs.get(k,0)
            xvec.append(x)
            yvec.append(y)

        cs = corsums(ranx(xvec),ranx(yvec))
        ##  cs = corsums(xvec,yvec)
        cs = corrval(cs)
        return  cs
    
    def votesim (self,cn,thisdoc,paradat):
        "voted similarity of model of category number cn to given document."

        if not hasattr(thisdoc,"elastab"):
            thisdoc.elastab = tokslabs(thisdoc.toklist,self.spansize,
                                 self.stopmode,self.spanmode,self.tops)
            thisdoc.elastot = sum(thisdoc.elastab.values())            
        votesum = 0.0
        divisor = 0
        fulltote = self.fulltote
        modtote = self.modlist[cn].elastots
        errmess = False
        ##  if Showoff > 0:
            ##  print(doctoks,modtoks,fulltoks,thisdoc.name,file=paradat.logfile)
        for fs in thisdoc.elastab:
            docfreq = thisdoc.elastab[fs]
            modfreq = self.modlist[cn].elastabs.get(fs,0)
            fullfreq = self.fulltab.get(fs,0)
            f0 = fullfreq
            n0 = fulltote
            f1 = modfreq ; n1 = modtote
            if f1 < 0 or n1 < 1 or f0 < 0 or n0 < 1:
                if not errmess:
                    print(f0,n0,f1,n1)
                    print("Trouble with counts in matchmod() !!")
                    errmess = True
                n1 = max(1,n1) ; n0 = max(1,n0)
            r0 = f0 / n0 ; r1 = f1 / n1  ##  rates
            ##  roar = (1.0-r0) / (1.0-r1)
            ##  vote = math.tanh(r1-r0)
            ##  vote = roar
            vote = 0.5 + 0.5*(r1>r0) - 0.5*(r0>r1)
            rf = docfreq ** 0.5
            ##  if self.pars.rootfrex > 0:
                ##  rf = rf ** 0.5
            ##  if self.pars.flatfrex > 0:  ##  takes precedence
                ##  rf = 1  ##  just count each token equally !
            votesum += (vote * rf)  ##  used 2b math.log(.)
            divisor += rf

        return  votesum/max(1,divisor)  ##  matching score, higher implies stronger match

    def matchmod (self,cn,thisdoc,paradat):
        "matches model for category number cn to given document."

        ##  s = self.votesim(cn,thisdoc,paradat)
        s = self.corrsim(cn,thisdoc,paradat)
        ##  s = self.attested(cn,thisdoc,paradat)
        ##  s = self.logbayes(cn,thisdoc,paradat)
        ##  s = random.random()
        
        return  s
    
    def modsims (self,thisdoc,paradat):
        "returns association score of each model to this doc."

        catvals = [0.0] * paradat.cats
        for cpos,c in enumerate(paradat.catlist):
            catvals[cpos] = self.matchmod(cpos,thisdoc,paradat)

        return  catvals
        ##  no scaling to quasi-probs at present.
        ##  return  [1.0/(1e-7+catvals[cpos]) for cpos in range(self.classes)]
        ##  inversion of distances used, so more means better.
